import os, platform
import boto3
import re

from typing import Optional, Tuple

from utils import is_valid_str_or_raise, compare_semver

REGION = "us-west-2"
BUCKET_NAME = "ikv-binaries"
SEMVER_PATTERN = r'^(\d+)\.(\d+)\.(\d+)(?:-([\w\.]+))?(?:\+([\w\.]+))?$'
CHUNK_SIZE = 8388608 # 8KB 

class NativeBinaryManager:
    """
    Responsible for dowloading native IKV binary, based on the host
    platform, for dynamic loading.
    """
    def __init__(self, mount_dir: str):
        self.s3_client = boto3.client('s3', region_name=REGION)
        self.mount_dir = is_valid_str_or_raise(mount_dir)
    
    def _init_mount_dir(self):
        """
        Ensures mount & mount/bin directories are usable.
        """
        # creates parent directories if required
        os.makedirs("{}/bin".format(self.mount_dir), exist_ok=True)

    def get_path_to_dll(self) -> Optional[str]:
        self._init_mount_dir()

        # ex. 0.0.3, /path/to/mount/bin/0.0.3-libikv.so
        local_semver, local_dll_path = self._local_dll_details()
        
        # ex. 0.0.5, release/mac-aarch64/0.0.5-libikv.dylib
        remote_semver, remote_dllfile_key = self._remote_dll_details()

        # nothing available on remote - return Optional local
        if remote_semver is None or remote_dllfile_key is None:
            return local_dll_path
        
        # local is present and is higher or equal as compared to remote
        if local_semver is not None and compare_semver(remote_semver, local_semver) <= 0:
            return local_dll_path
        
        # download remote
        return self._download_remote_dll(remote_dllfile_key)

    def _local_dll_details(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Local dll details if it exists (or None, None).
        Return: semver, path tuple, ex. (0.0.3, ..mount-directory/bin/0.0.3-libikv.so)
        """
        path = "{}/bin".format(self.mount_dir)
        filenames = os.listdir(path=path)
        if not filenames: 
            return None, None
        
        for filename in filenames:
            # filename: 0.0.3-libikv.so
            semver = NativeBinaryManager.parse_semver(filename)
            if semver is not None:
                return semver, "{}/bin/{}".format(self.mount_dir, filename)
            
        return None, None
    
    def _remote_dll_details(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Remote dll details (highest available semver) if it exists (or None, None).
        Return: semver, filename tuple, ex. (0.0.5, release/mac-aarch64/0.0.5-libikv.dylib)
        """
        # release/{mac|linux|windows}-{aarch64|x86_64|tbd}
        platform_prefix = NativeBinaryManager.platform_s3_prefix()
        if platform_prefix is None:
            return None, None

        response = self.s3_client.list_objects_v2(Bucket=BUCKET_NAME, Prefix=platform_prefix)
        
        remote_files = []
        if 'Contents' in response:
            objects = [obj for obj in response['Contents'] if obj is not None]
            keys = [obj['Key'] for obj in objects if obj['Key'] is not None]
            remote_files = [key.split("/")[-1] for key in keys] # output: ["foo", "0.0.3-libikv.so", "bar"]
            remote_files = [f for f in remote_files if NativeBinaryManager.parse_semver(f) is not None] # output: ["0.0.3-libikv.so", "0.0.5-libikv.so"]

        if len(remote_files) == 0:
            return None, None

        # pick highest semver
        highest_semver = NativeBinaryManager.parse_semver(remote_files[0])
        highest_semver_file = remote_files[0]
        for file in remote_files:
            if compare_semver(NativeBinaryManager.parse_semver(file), highest_semver) > 0:
                highest_semver = NativeBinaryManager.parse_semver(file)
                highest_semver_file = file

        return (highest_semver, "{}/{}".format(platform_prefix, highest_semver_file))
    
    def _download_remote_dll(self, object_key: str) -> Optional[str]:
        """
        Returns full file path to local bin after downloading it from remote.
        Also deletes older local binaries.
        """
        # delete existing binaries
        NativeBinaryManager._delete_files_in_directory("{}/bin".format(self.mount_dir))

        # ex. release/mac-aarch64/0.0.5-libiky.dylib -> 0.0.5-libiky.dylib
        remote_dll_filename = object_key.split("/")[-1]
        # ex. /path/to/mount/bin/0.0.5-libiky.dylib
        local_dll_path = "{}/bin/{}".format(self.mount_dir, remote_dll_filename)

        # download from s3 in chunks
        response = self.s3_client.get_object(Bucket=BUCKET_NAME, Key=object_key)
        if response is None or response['Body'] is None:
            return None

        with open(local_dll_path, 'wb') as local_dll_file:
            for chunk in response['Body'].iter_chunks(CHUNK_SIZE):
                local_dll_file.write(chunk)
        
        return local_dll_path

    def _delete_files_in_directory(directory):
        for file_name in os.listdir(directory):
            file_path = os.path.join(directory, file_name)
            if os.path.isfile(file_path):
                os.remove(file_path)

    def parse_semver(filename: str) -> Optional[str]:
        # input: 0.0.3-libikv.so, output: 0.0.3
        # input: 0.0.4, output: 0.0.4
        # input: libikv.so, output: None
        if len(filename) == 0: return None
        
        parts = filename.split("-")
        if not parts: return None
        
        if re.match(SEMVER_PATTERN, parts[0]) is not None: return parts[0]
        return None

    def platform_s3_prefix() -> Optional[str]:
        # return: release/{mac|linux|windows}-{aarch64|x86_64|tbd}
        raw_platform_os = str(platform.system()).lower()
        raw_platform_machine = str(platform.machine()).lower()

        if raw_platform_os == "linux":
            if raw_platform_machine == "x86_64" or raw_platform_machine == "amd64":
                return "release/linux-x86_64"
            elif raw_platform_machine == "arm64" or raw_platform_machine == "aarch64":
                return "release/linux-aarch64"
        elif raw_platform_os == "darwin" or raw_platform_os == "mac":
            if raw_platform_machine == "arm64" or raw_platform_machine == "aarch64":
                return "release/mac-aarch64"
        
        return None


from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class FieldType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN: _ClassVar[FieldType]
    INT32: _ClassVar[FieldType]
    INT64: _ClassVar[FieldType]
    FLOAT32: _ClassVar[FieldType]
    FLOAT64: _ClassVar[FieldType]
    STRING: _ClassVar[FieldType]
    BYTES: _ClassVar[FieldType]
    BOOLEAN: _ClassVar[FieldType]
UNKNOWN: FieldType
INT32: FieldType
INT64: FieldType
FLOAT32: FieldType
FLOAT64: FieldType
STRING: FieldType
BYTES: FieldType
BOOLEAN: FieldType

class IKVDocumentOnWire(_message.Message):
    __slots__ = ("document",)
    class DocumentEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: FieldValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[FieldValue, _Mapping]] = ...) -> None: ...
    DOCUMENT_FIELD_NUMBER: _ClassVar[int]
    document: _containers.MessageMap[str, FieldValue]
    def __init__(self, document: _Optional[_Mapping[str, FieldValue]] = ...) -> None: ...

class FieldValue(_message.Message):
    __slots__ = ("fieldType", "value")
    FIELDTYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    fieldType: FieldType
    value: bytes
    def __init__(self, fieldType: _Optional[_Union[FieldType, str]] = ..., value: _Optional[bytes] = ...) -> None: ...

class IKVStoreConfig(_message.Message):
    __slots__ = ("stringConfigs", "intConfigs", "floatConfigs", "bytesConfigs", "booleanConfigs")
    class StringConfigsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class IntConfigsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    class FloatConfigsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
    class BytesConfigsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bytes
        def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ...) -> None: ...
    class BooleanConfigsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    STRINGCONFIGS_FIELD_NUMBER: _ClassVar[int]
    INTCONFIGS_FIELD_NUMBER: _ClassVar[int]
    FLOATCONFIGS_FIELD_NUMBER: _ClassVar[int]
    BYTESCONFIGS_FIELD_NUMBER: _ClassVar[int]
    BOOLEANCONFIGS_FIELD_NUMBER: _ClassVar[int]
    stringConfigs: _containers.ScalarMap[str, str]
    intConfigs: _containers.ScalarMap[str, int]
    floatConfigs: _containers.ScalarMap[str, float]
    bytesConfigs: _containers.ScalarMap[str, bytes]
    booleanConfigs: _containers.ScalarMap[str, bool]
    def __init__(self, stringConfigs: _Optional[_Mapping[str, str]] = ..., intConfigs: _Optional[_Mapping[str, int]] = ..., floatConfigs: _Optional[_Mapping[str, float]] = ..., bytesConfigs: _Optional[_Mapping[str, bytes]] = ..., booleanConfigs: _Optional[_Mapping[str, bool]] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2
import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class HelloWorldRequest(_message.Message):
    __slots__ = ("echo",)
    ECHO_FIELD_NUMBER: _ClassVar[int]
    echo: str
    def __init__(self, echo: _Optional[str] = ...) -> None: ...

class HelloWorldResponse(_message.Message):
    __slots__ = ("echo",)
    ECHO_FIELD_NUMBER: _ClassVar[int]
    echo: str
    def __init__(self, echo: _Optional[str] = ...) -> None: ...

class HealthCheckResponse(_message.Message):
    __slots__ = ("code",)
    CODE_FIELD_NUMBER: _ClassVar[int]
    code: int
    def __init__(self, code: _Optional[int] = ...) -> None: ...

class UpsertFieldValuesRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "document")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENT_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    document: _common_pb2.IKVDocumentOnWire
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., document: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ...) -> None: ...

class BatchUpsertFieldValuesRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "documents")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTS_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    documents: _containers.RepeatedCompositeFieldContainer[_common_pb2.IKVDocumentOnWire]
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., documents: _Optional[_Iterable[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]]] = ...) -> None: ...

class DeleteFieldValueRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "documentId", "fieldNames")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTID_FIELD_NUMBER: _ClassVar[int]
    FIELDNAMES_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    documentId: _common_pb2.IKVDocumentOnWire
    fieldNames: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., documentId: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ..., fieldNames: _Optional[_Iterable[str]] = ...) -> None: ...

class BatchDeleteFieldValuesRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "documentIds", "fieldNames")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTIDS_FIELD_NUMBER: _ClassVar[int]
    FIELDNAMES_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    documentIds: _containers.RepeatedCompositeFieldContainer[_common_pb2.IKVDocumentOnWire]
    fieldNames: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., documentIds: _Optional[_Iterable[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]]] = ..., fieldNames: _Optional[_Iterable[str]] = ...) -> None: ...

class DeleteDocumentRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "documentId")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTID_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    documentId: _common_pb2.IKVDocumentOnWire
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., documentId: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ...) -> None: ...

class BatchDeleteDocumentsRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "documentIds")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTIDS_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    documentIds: _containers.RepeatedCompositeFieldContainer[_common_pb2.IKVDocumentOnWire]
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., documentIds: _Optional[_Iterable[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]]] = ...) -> None: ...

class DropFieldsRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer", "timestamp", "field_names", "field_name_prefixes", "drop_all")
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    FIELD_NAMES_FIELD_NUMBER: _ClassVar[int]
    FIELD_NAME_PREFIXES_FIELD_NUMBER: _ClassVar[int]
    DROP_ALL_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    timestamp: _timestamp_pb2.Timestamp
    field_names: _containers.RepeatedScalarFieldContainer[str]
    field_name_prefixes: _containers.RepeatedScalarFieldContainer[str]
    drop_all: bool
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., field_names: _Optional[_Iterable[str]] = ..., field_name_prefixes: _Optional[_Iterable[str]] = ..., drop_all: bool = ...) -> None: ...

class FieldSchema(_message.Message):
    __slots__ = ("name", "id", "fieldType")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    FIELDTYPE_FIELD_NUMBER: _ClassVar[int]
    name: str
    id: int
    fieldType: _common_pb2.FieldType
    def __init__(self, name: _Optional[str] = ..., id: _Optional[int] = ..., fieldType: _Optional[_Union[_common_pb2.FieldType, str]] = ...) -> None: ...

class GetUserStoreConfigRequest(_message.Message):
    __slots__ = ("userStoreContextInitializer",)
    USERSTORECONTEXTINITIALIZER_FIELD_NUMBER: _ClassVar[int]
    userStoreContextInitializer: UserStoreContextInitializer
    def __init__(self, userStoreContextInitializer: _Optional[_Union[UserStoreContextInitializer, _Mapping]] = ...) -> None: ...

class GetUserStoreConfigResponse(_message.Message):
    __slots__ = ("globalConfig",)
    GLOBALCONFIG_FIELD_NUMBER: _ClassVar[int]
    globalConfig: _common_pb2.IKVStoreConfig
    def __init__(self, globalConfig: _Optional[_Union[_common_pb2.IKVStoreConfig, _Mapping]] = ...) -> None: ...

class Status(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UserStoreContextInitializer(_message.Message):
    __slots__ = ("credentials", "storeName")
    CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
    STORENAME_FIELD_NUMBER: _ClassVar[int]
    credentials: AccountCredentials
    storeName: str
    def __init__(self, credentials: _Optional[_Union[AccountCredentials, _Mapping]] = ..., storeName: _Optional[str] = ...) -> None: ...

class AccountCredentials(_message.Message):
    __slots__ = ("accountId", "accountPasskey")
    ACCOUNTID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTPASSKEY_FIELD_NUMBER: _ClassVar[int]
    accountId: str
    accountPasskey: str
    def __init__(self, accountId: _Optional[str] = ..., accountPasskey: _Optional[str] = ...) -> None: ...



from typing import Dict
from schemas.common_pb2 import FieldValue, IKVDocumentOnWire
from schemas.common_pb2 import STRING as FieldValueString
from schemas.common_pb2 import BYTES as FieldValueBytes
import utils as ikvutils

class IKVDocument:
    """ 
    Representation of documents for writing into IKV. 
    See IKVDocumentBuilder to instantiate. 
    """
    def __init__(self, fields: Dict[str, FieldValue]):
        """ DO NOT use directly, see IKVDocumentBuilder to instantiate."""
        if fields is None:
            raise ValueError("fields cannot be None")
        self.fields: Dict[str, FieldValue] = fields
    
    def len(self):
        return len(self.fields)
    
    def as_ikv_document_on_wire(self) -> IKVDocumentOnWire:
        return IKVDocumentOnWire(document=self.fields)

class IKVDocumentBuilder:
    def __init__(self):
        self.fields: Dict[str, FieldValue] = {}
    
    def put_string_field(self, name: str, value: str) -> 'IKVDocumentBuilder':
        name: str = ikvutils.is_valid_str_or_raise(name)
        value: bytes = ikvutils.is_valid_str_or_raise(value).encode('utf-8')

        field_value: FieldValue = FieldValue(fieldType=FieldValueString, value=value)
        self.fields[name] = field_value

        return self
    
    def put_bytes_field(self, name: str, value: bytes) -> 'IKVDocumentBuilder':
        name: str = ikvutils.is_valid_str_or_raise(name)
        value: bytes = ikvutils.is_valid_bytes_or_raise(value)

        field_value: FieldValue = FieldValue(fieldType=FieldValueBytes, value=value)
        self.fields[name] = field_value
        
        return self

    def build(self) -> IKVDocument:
        return IKVDocument(fields=self.fields)


from typing import Iterator, List, Optional
from cffi import FFI
import utils
from utils import is_valid_str_or_raise, is_valid_bytes_or_raise

EMPTY_ITERATOR = iter([])

ffi = FFI()
ffi.cdef("""
    // Start of common C code (Go, Python)
    typedef struct BytesBuffer {
    int32_t length;
    uint8_t *start;
    } BytesBuffer;

    typedef struct IndexHandle {
    int64_t handle;
    int64_t status;
    } IndexHandle;

    int64_t health_check(const char *input);
    
    IndexHandle open_index_v2(const char *config, int32_t config_len);
    
    void close_index(int64_t handle);
    
    BytesBuffer get_field_value(int64_t handle, const char *pkey, int32_t pkey_len, const char *field_name);

    BytesBuffer multiget_field_values(int64_t handle, const char *pkeys, int32_t pkeys_len, const char *field_names, int32_t field_names_len);

    void free_bytes_buffer(BytesBuffer buf);
    // End of common C code (Go, Python)
""")

class NativeReader:
    def __init__(self, dll_path: str):
        self.index_handle = -1
        self.dll = ffi.dlopen(dll_path)
    
    def health_check(self, input: str) -> bool:
        # this will be auto free'd with python gc
        c_input = ffi.new("char[]", is_valid_str_or_raise(input).encode('utf-8'))
        status = self.dll.health_check(c_input)
        return status == 0

    def open(self, ikv_config: bytes):
        if self.index_handle != -1:
            return

        # this will be auto free'd with python gc
        c_ikv_config = ffi.new("char[]", is_valid_bytes_or_raise(ikv_config))
        index_handle = self.dll.open_index_v2(c_ikv_config, len(ikv_config))

        if index_handle.status == 0:
            self.index_handle = index_handle.handle
            return
        
        # error
        raise RuntimeError("Cannot open IKV reader, error code: {}".format(index_handle.status))

    def close(self):
        if self.index_handle == -1:
            return

        self.dll.close_index(self.index_handle)

    def get_bytes_field_value(self, primary_key: bytes, field_name: str) -> Optional[bytes]:
        c_primary_key = ffi.from_buffer(primary_key)
        c_field_name = ffi.from_buffer(field_name.encode('utf-8'))

        bytes_buffer = self.dll.get_field_value(self.index_handle, c_primary_key, len(primary_key), c_field_name)
        
        value_len = bytes_buffer.length
        value_start = bytes_buffer.start

        if value_len == 0 or value_start == ffi.NULL:
            return None
        
        # copy from the pointer
        value = bytes(ffi.buffer(value_start, value_len))

        # release rust allocated objects
        self.dll.free_bytes_buffer(bytes_buffer)

        return value

    def get_string_field_value(self, primary_key: bytes, field_name: str) -> Optional[str]:
        c_primary_key = ffi.from_buffer(primary_key)
        c_field_name = ffi.from_buffer(field_name.encode('utf-8'))

        bytes_buffer = self.dll.get_field_value(self.index_handle, c_primary_key, len(primary_key), c_field_name)

        value_len = bytes_buffer.length
        value_start = bytes_buffer.start

        if value_len == 0 or value_start == ffi.NULL:
            return None

        # zero copy decoding into str
        value = str(memoryview(ffi.buffer(value_start, value_len)), 'utf-8')

        # release rust allocated objects
        self.dll.free_bytes_buffer(bytes_buffer)

        return value

    def multiget_bytes_field_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],
            field_names: List[str] = []) -> Iterator[Optional[bytes]]:
        """ native muti-get wrapper. """

        num_values = (len(bytes_primary_keys) + len(str_primary_keys)) * len(field_names)
        if num_values == 0:
            return EMPTY_ITERATOR

        concat_primary_keys: bytearray = utils.concat_as_utf8_with_size_prefix(bytes_primary_keys, str_primary_keys)
        c_concat_primary_keys = ffi.from_buffer(concat_primary_keys)
        concat_field_names: bytearray = utils.concat_strings_with_size_prefix(field_names)
        c_concat_field_names = ffi.from_buffer(concat_field_names)

        bytes_buffer = self.dll.multiget_field_values(self.index_handle, c_concat_primary_keys,
            len(concat_primary_keys), c_concat_field_names, len(concat_field_names))

        value_len = bytes_buffer.length
        value_start = bytes_buffer.start
        if value_len == 0 or value_start == ffi.NULL:
            return iter([None] * num_values)

        # copy from the pointer
        value = bytes(ffi.buffer(value_start, value_len))

        # release rust allocated objects
        self.dll.free_bytes_buffer(bytes_buffer)

        # return iterator over `value`
        return utils.unpack_size_prefixed_bytes_as_bytes(value)

    def multiget_str_field_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],
            field_names: List[str] = []) -> Iterator[Optional[str]]:
        """ native muti-get wrapper. """

        num_values = (len(bytes_primary_keys) + len(str_primary_keys)) * len(field_names)
        if num_values == 0:
            return EMPTY_ITERATOR

        concat_primary_keys: bytearray = utils.concat_as_utf8_with_size_prefix(bytes_primary_keys, str_primary_keys)
        c_concat_primary_keys = ffi.from_buffer(concat_primary_keys)
        concat_field_names: bytearray = utils.concat_strings_with_size_prefix(field_names)
        c_concat_field_names = ffi.from_buffer(concat_field_names)

        bytes_buffer = self.dll.multiget_field_values(self.index_handle, c_concat_primary_keys,
            len(concat_primary_keys), c_concat_field_names, len(concat_field_names))

        value_len = bytes_buffer.length
        value_start = bytes_buffer.start
        if value_len == 0 or value_start == ffi.NULL:
            return iter([None] * num_values)

        # copy from the pointer
        value = bytes(ffi.buffer(value_start, value_len))

        # release rust allocated objects
        self.dll.free_bytes_buffer(bytes_buffer)

        # return iterator over `value`
        return utils.unpack_size_prefixed_bytes_as_str(value)


import unittest
from src.ikvpy.native_reader import NativeReader
from src.ikvpy.clientoptions import ClientOptions, ClientOptionsBuilder
import src.ikvpy.client as ikv_client
from src.ikvpy.factory import create_new_reader
from env_var import EnvReader

class TestIKVReaderImpl(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # setup client options
        env = EnvReader()
        client_options: ClientOptions = ClientOptionsBuilder() \
            .with_account_id(env.get("ACCOUNT_ID"))\
            .with_account_passkey(env.get("ACCOUNT_PASSKEY"))\
            .with_store_name(env.get("STORE_NAME"))\
            .with_mount_directory(env.get("MOUNT_DIR"))\
            .build()
        
        # setup reader
        self.reader: ikv_client.IKVReader = create_new_reader(client_options)

    def test_startup(self):
        self.reader.startup()
    
    def test_shutdown(self):
        self.reader.shutdown()

# run: $> python3 tests/test_native_reader.py
#      $> [ikv-python-client/tests]$ python3 -m unittest
if __name__ == '__main__':
    unittest.main()


import struct

from typing import List, Tuple, Optional, Iterator

def is_valid_str_or_raise(input: str) -> str:
    if not isinstance(input, str):
        raise TypeError("expecting a string")
    if not input.strip():
        raise ValueError("expecting non-empty string")
    return input

def is_valid_bytes_or_raise(input: bytes) -> bytes:
    if not isinstance(input, bytes):
        raise TypeError("expecting bytes")
    if len(input) == 0:
        raise ValueError("expecting non-empty bytes")
    return input

def _parse_semver(version) -> Tuple[int, int, int, Optional[str]]:
    version_parts = version.split('.')
    major, minor, patch = map(int, version_parts[:3])
    pre_release = version_parts[3] if len(version_parts) > 3 and '-' in version_parts[3] else None
    return major, minor, patch, pre_release

def compare_semver(version1, version2):
    """
    Compare two semver strings.
    result > 0 iff v1 > v2
    result < 0 iff v1 < v2
    result == 0 iff v1 == v2
    """
    v1 = _parse_semver(version1)
    v2 = _parse_semver(version2)
    return (v1 > v2) - (v1 < v2)

def concat_strings_with_size_prefix(input: List[str] = []) -> bytearray:
    if not len(input):
        return bytearray()

    capacity = 0
    input_as_bytes = [None] * len(input)
    for i, inp in enumerate(input):
        inp_bytes = inp.encode("utf-8")
        capacity += 4 + len(inp_bytes)
        input_as_bytes[i] = inp_bytes

    output = bytearray(capacity)

    i = 0
    for inp in input_as_bytes:
        # lower endian i32
        struct.pack_into('<i', output, i, len(inp))
        i += 4
        # utf8 bytes
        output[i:i+len(inp)] = inp
        i += len(inp)

    return output

def concat_as_utf8_with_size_prefix(bytes_input: List[bytes] = [], str_input: List[str] = []) -> bytearray:
    if not len(bytes_input) and not len(str_input):
        return bytearray()

    capacity = 0
    for inp in bytes_input:
        capacity += 4 + len(inp)

    str_input_as_bytes = [None] * len(str_input)
    for i, inp in enumerate(str_input):
        inp_bytes = inp.encode("utf-8")
        capacity += 4 + len(inp_bytes)
        str_input_as_bytes[i] = inp_bytes

    output = bytearray(capacity)

    i = 0
    for inp_list in [bytes_input, str_input_as_bytes]:
        for inp in inp_list:
            # lower endian i32
            struct.pack_into('<i', output, i, len(inp))
            i += 4
            # utf8 bytes
            output[i:i+len(inp)] = inp
            i += len(inp)

    return output

def unpack_size_prefixed_bytes_as_bytes(input: bytes) -> Iterator[Optional[bytes]]:
    i = 0
    while i < len(input):
        size = struct.unpack_from('<i', input, i)[0]
        i += 4

        if size == -1:
            yield None
        elif size == 0:
            yield bytes()
        else:
            yield input[i:i+size]
            i += size

def unpack_size_prefixed_bytes_as_str(input: bytes) -> Iterator[Optional[str]]:
    i = 0
    while i < len(input):
        size = struct.unpack_from('<i', input, i)[0]
        i += 4

        if size == -1:
            yield None
        elif size == 0:
            yield ""
        else:
            yield str(memoryview(input[i:i+size]), 'utf-8')
            i += size




from abc import ABC, abstractmethod
from clientoptions import ClientOptions
from document import IKVDocument
from typing import Iterator, List, Optional

class IKVReader(ABC):
    """
    See https://docs.inlined.io for detailed Python usage guide.
    See factory.py for instantiating a concrete reader instance.

    Reader client over a provisioned IKV store.
    You should reuse an instantiated client as it holds an embedded database instance.
    Lifecycle (startup/shutdown) of a reader client should be tied with application lifecycle.
    """
    @abstractmethod
    def startup(self):
        """
        Blocking startup, initializes the embedded store by pulling latest data locally.
        Do not invoke other functions before startup is finished.
        """
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")

    @abstractmethod
    def shutdown(self):
        """
        Shutdown embedded store - ex. during application exit.
        """
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")

    @abstractmethod
    def get_bytes_value(self, primary_key, field_name: str) -> Optional[bytes]:
        """
        Fetch an inner field of type bytes, by providing the primary-key for the document and the field-name.
        Please note IKV readers are eventually-consistent in terms of "reading your writes".
        There can be a very small delay between writing document (with IKVWriter) and reading them.

        args:
            primary_key: document identifier (bytes or str)
            field_name: field identifier
        returns:
            field value if it exists (bytes or None)
        """
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")

    @abstractmethod
    def multiget_bytes_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],\
            field_names: List[str] = []) -> Iterator[Optional[bytes]]:
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")

    @abstractmethod
    def multiget_string_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],\
            field_names: List[str] = []) -> Iterator[Optional[str]]:
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")

    @abstractmethod
    def get_string_value(self, primary_key, field_name: str) -> Optional[str]:
        """
        Fetch an inner field of type str, by providing the primary-key for the document and the field-name.
        Please note IKV readers are eventually-consistent in terms of "reading your writes".
        There can be a very small delay between writing document (with IKVWriter) and reading them.

        args:
            primary_key: document identifier (bytes or str)
            field_name: field identifier
        returns:
            field value if it exists (str or None)
        """
        raise NotImplementedError("cannot be used directly, use IKVReaderImpl")


class IKVWriter(ABC):
    """
    See https://docs.inlined.io for detailed Python usage guide.
    See factory.py for instantiating a concrete writer instance.

    Writer client over a provisioned IKV store.
    You should reuse an instantiated client.
    Lifecycle (startup/shutdown) of a writer client should be tied with application lifecycle.
    """
    @abstractmethod
    def startup(self):
        """
        Initializes the writer client. Do not invoke other functions before startup is finished.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")

    @abstractmethod
    def shutdown(self):
        """
        Shutdown writer client - ex. during application exit.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")
    
    @abstractmethod
    def upsert_fields(self, document: IKVDocument):
        """
        Upsert (insert or update) fields for a document.
        Provided `document` must contain the value of the primary-key and (if-applicable) partitioning-key.
        See document.py for creating IKVDocument objects.

        Different values for the same primary-key are aggregated by unionizing distincts and overwriting duplicates. Example -
        upsert: {"pkey": 0, "key1": "value1"}, followed by upsert: {"pkey": 0, "key2": "value2"}
        results in {"pkey": 0, "key1": "value1", "key2": "value2"} being saved.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")

    @abstractmethod
    def delete_fields(self, document: IKVDocument, fields_to_delete: List[str]):
        """
        Delete specified fields from a document. NO OP if document or any field in `fields_to_delete` does not exist.
        Provided `document` must contain the value of the primary-key and (if-applicable) partitioning-key.
        See document.py for creating IKVDocument objects.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")

    @abstractmethod
    def delete_document(self, document: IKVDocument):
        """
        Delete a document (i.e. all of its fields) if it exists. NO OP if `document` is not indexed.
        Provided `document` must contain the value of the primary-key and (if-applicable) partitioning-key.
        See document.py for creating IKVDocument objects.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")

    @abstractmethod
    def drop_fields_by_name(self, field_names: List[str]):
        """
        Drop specified fields for all documents.
        Attempts to drop primary-key field are silently ignored (no error).
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")

    @abstractmethod
    def drop_fields_by_name_prefix(self, field_name_prefixes: List[str]):
        """
        Drop specified fields for all documents, by specifying field name prefixes.
        Attempts to drop primary-key field are silently ignored (no error).
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")
    
    @abstractmethod
    def drop_all_fields(self):
        """
        Drop all fields (except primary-key) for all documents.
        """
        raise NotImplementedError("cannot be used directly, use IKVWriterImpl")


import unittest
import src.ikvpy.utils as utils

class TestUtils(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    def test_roundtrip_strings(self):
        cases = [["foo", "bar", "", "bar"], [], [""]]
        
        # write as str, read as str
        for input in cases:        
            serialized = bytes(utils.concat_strings_with_size_prefix(input))
            deserialized = list(utils.unpack_size_prefixed_bytes_as_str(serialized))
            self.assertEqual(input, deserialized)
        
        # write as str, read as bytes
        for input in cases:        
            serialized = bytes(utils.concat_as_utf8_with_size_prefix([], input))
            deserialized = list(utils.unpack_size_prefixed_bytes_as_bytes(serialized))
            deserialized = [str(x, "utf-8") for x in deserialized]
            self.assertEqual(input, deserialized)

    def test_roundtrip_bytes(self):
        cases = [["foo", "bar", "", "bar"], [], [""]]
        
        # write as bytes, read as bytes
        for input in cases:
            input = [x.encode("utf-8") for x in input]     
            serialized = bytes(utils.concat_as_utf8_with_size_prefix(input, []))
            deserialized = list(utils.unpack_size_prefixed_bytes_as_bytes(serialized))
            self.assertEqual(input, deserialized)

    def test_roundtrip(self):
        input_bytes = ["bar".encode(), "".encode()]
        input_str = ["", "foo"]

        serialized = bytes(utils.concat_as_utf8_with_size_prefix(input_bytes, input_str))
        deserialized = list(utils.unpack_size_prefixed_bytes_as_bytes(serialized))

        self.assertEqual(input_bytes[0], deserialized[0])
        self.assertEqual(input_bytes[1], deserialized[1])
        self.assertEqual(input_str[0], str(deserialized[2], "utf-8"))
        self.assertEqual(input_str[1], str(deserialized[3], "utf-8"))

# run: $> python3 tests/test_utils.py
#      $> [ikv-python-client/tests]$ python3 -m unittest
if __name__ == '__main__':
    unittest.main()


import json
import grpc
import schemas.services_pb2_grpc as services_pb2_grpc
import schemas.services_pb2 as services_pb2
import schemas.common_pb2 as common_pb2

from google.protobuf import timestamp_pb2
from grpc_status import rpc_status

from client import IKVWriter
from document import IKVDocument
from typing import List, Optional
from clientoptions import ClientOptions
from utils import is_valid_str_or_raise

# grpc call retry policy
retry_policy = json.dumps(
    {
        "methodConfig": [{
            "name": [{"service": "ikvschemas.InlineKVWriteService"}],
            "retryPolicy": {
			  "maxAttempts": 3,
			  "initialBackoff": "0.01s",
			  "maxBackoff": "0.01s",
			  "backoffMultiplier": 1.0,
			  "retryableStatusCodes": [ "UNAVAILABLE" ]
            },
        }]
    }
)

class IKVWriterImpl(IKVWriter):
    def __init__(self, client_options: ClientOptions):
        if client_options is None:
            raise TypeError("client_options are required and can't be None")
        self.client_options = client_options
        
        # initialize auth context initializer
        account_id = is_valid_str_or_raise(client_options.get_ikv_config().stringConfigs["account_id"])
        account_passkey = is_valid_str_or_raise(client_options.get_ikv_config().stringConfigs["account_passkey"])
        credentials: services_pb2.AccountCredentials = \
            services_pb2.AccountCredentials(accountId=account_id, accountPasskey=account_passkey)
        store_name = is_valid_str_or_raise(client_options.get_ikv_config().stringConfigs["store_name"])
        self.user_store_context_initializer: services_pb2.UserStoreContextInitializer = \
            services_pb2.UserStoreContextInitializer(credentials=credentials, storeName=store_name)

        self.channel = None
        self.stub = None

    def startup(self):
        if self.channel is None:
            options = []
            options.append(("grpc.service_config", retry_policy))
            self.channel = grpc.secure_channel(target="gateway.inlined.io:443", credentials=grpc.ssl_channel_credentials(), options=options)
            self.stub = services_pb2_grpc.InlineKVWriteServiceStub(self.channel)

    def shutdown(self):
        if self.channel is not None:
            self.channel.close()
            self.channel = None
            self.stub = None

    def fetch_server_supplied_config(self) -> Optional[common_pb2.IKVStoreConfig]:
        request: services_pb2.GetUserStoreConfigRequest = services_pb2.GetUserStoreConfigRequest(\
            userStoreContextInitializer=self.user_store_context_initializer)

        try:
            response: services_pb2.GetUserStoreConfigResponse = self.stub.getUserStoreConfig(request)
            return response.globalConfig
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def upsert_fields(self, document: IKVDocument):
        if document is None or document.len() < 1:
            raise ValueError("Not a valid document, must contain primary-key")

        request = services_pb2.UpsertFieldValuesRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), \
                document=document.as_ikv_document_on_wire())
        
        try:
            self.stub.upsertFieldValues(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def delete_fields(self, document: IKVDocument, fields_to_delete: List[str]):
        if document is None or document.len() < 1:
            raise ValueError("Not a valid document, must contain primary-key")
        
        if fields_to_delete is None or len(fields_to_delete) == 0:
            return

        request = services_pb2.DeleteFieldValueRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), \
                documentId=document.as_ikv_document_on_wire(), fieldNames=fields_to_delete)

        try:
            self.stub.deleteFieldValues(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def delete_document(self, document: IKVDocument):
        if document is None or document.len() < 1:
            raise ValueError("Not a valid document, must contain primary-key")
        
        request = services_pb2.DeleteDocumentRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), \
                documentId=document.as_ikv_document_on_wire())

        try:
            self.stub.deleteDocument(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def drop_fields_by_name(self, field_names: List[str]):
        if field_names is None or len(field_names) == 0:
            return

        request: services_pb2.DropFieldsRequest = services_pb2.DropFieldsRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), \
                field_names=field_names, drop_all=False)
        
        try:
            self.stub.dropFields(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def drop_fields_by_name_prefix(self, field_name_prefixes: List[str]):
        if field_name_prefixes is None or len(field_name_prefixes) == 0:
            return

        request: services_pb2.DropFieldsRequest = services_pb2.DropFieldsRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), \
                field_name_prefixes=field_name_prefixes, drop_all=False)
        
        try:
            self.stub.dropFields(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")

    def drop_all_fields(self):
        request: services_pb2.DropFieldsRequest = services_pb2.DropFieldsRequest(\
            userStoreContextInitializer=self.user_store_context_initializer, timestamp=self._create_timestamp(), drop_all=True)
        
        try:
            self.stub.dropFields(request)
        except grpc.RpcError as rpc_error:
            status = rpc_status.from_call(rpc_error)
            raise RuntimeError(f"Unexpected failure, status code: {status.code} message: {status.message}")
    
    def _create_timestamp(self):
        timestamp = timestamp_pb2.Timestamp()
        timestamp.GetCurrentTime()
        return timestamp


from typing import List, Optional, Iterator
from client import IKVReader
from clientoptions import ClientOptions
from writer import IKVWriterImpl
from bin_manager import NativeBinaryManager
from utils import is_valid_str_or_raise

import native_reader
import schemas.common_pb2 as common_pb2

class IKVReaderImpl(IKVReader):
    def __init__(self, client_options: ClientOptions):
        """ See client.py for usage documentation. """
        if client_options is None:
            raise TypeError("client_options are required and can't be None")
        self.client_options = client_options
        self.native_reader = None

    def startup(self):
        """ See client.py for usage documentation. """
        # download dll and initialize native reader
        mount_dir = is_valid_str_or_raise(self.client_options.get_ikv_config().stringConfigs["mount_directory"])
        bin_manager = NativeBinaryManager(mount_dir=mount_dir)
        dll_path = bin_manager.get_path_to_dll()
        if dll_path is None: raise RuntimeError("Cannot download IKV native binary")
        self.native_reader = native_reader.NativeReader(dll_path)

        # fetch server supplied config and merge with client cfg
        writer: IKVWriterImpl = IKVWriterImpl(self.client_options)
        writer.startup()
        server_cfg = writer.fetch_server_supplied_config()
        writer.shutdown()
        if server_cfg is None: raise RuntimeError("cannot fetch startup cfg from IKV cloud for given client options")
        merged_cfg = self._merge_configs(server_cfg)
        ikv_config_bytes = bytes(merged_cfg.SerializeToString())
        
        # open embedded db
        self.native_reader.open(ikv_config_bytes)

    def shutdown(self):
        self.native_reader.close()

    def get_bytes_value(self, primary_key, field_name: str) -> Optional[bytes]:
        """ See client.py for usage documentation. """
        if isinstance(primary_key, str):
            return self.native_reader.get_bytes_field_value(bytes(primary_key.encode('utf-8')), field_name)
        
        if isinstance(primary_key, bytes):
            return self.native_reader.get_bytes_field_value(primary_key, field_name)
        
        # ffi.new() does not accept bytearray type as is
        if isinstance(primary_key, bytearray):
            return self.native_reader.get_bytes_field_value(bytes(primary_key), field_name)
        
        raise TypeError("unsupported primary_key type: {}, supported: str/bytes/bytearray".format(type(primary_key)))

    def get_string_value(self, primary_key, field_name: str) -> Optional[str]:
        """ See client.py for usage documentation. """
        if isinstance(primary_key, str):
            return self.native_reader.get_string_field_value(bytes(primary_key.encode('utf-8')), field_name)
        
        if isinstance(primary_key, bytes):
            return self.native_reader.get_string_field_value(primary_key, field_name)
        
        # ffi.new() does not accept bytearray type as is
        if isinstance(primary_key, bytearray):
            return self.native_reader.get_string_field_value(bytes(primary_key), field_name)

        raise TypeError("unsupported primary_key type: {}, supported: str/bytes/bytearray".format(type(primary_key)))

    def multiget_bytes_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],\
            field_names: List[str] = []) -> Iterator[Optional[bytes]]:
        return self.native_reader.multiget_bytes_field_values(bytes_primary_keys, str_primary_keys, field_names)

    def multiget_string_values(self, bytes_primary_keys: List[bytes] = [], str_primary_keys: List[str] = [],\
            field_names: List[str] = []) -> Iterator[Optional[str]]:
        return self.native_reader.multiget_str_field_values(bytes_primary_keys, str_primary_keys, field_names)

    def _merge_configs(self, server_cfg: common_pb2.IKVStoreConfig) -> common_pb2.IKVStoreConfig:
        """ Internal method, overrides & merges server supplied cfg with client supplier cfg. """
        client_cfg = self.client_options.get_ikv_config()

        string_configs = server_cfg.stringConfigs if server_cfg.stringConfigs is not None else {}
        if client_cfg.stringConfigs is not None:
            for k,v in client_cfg.stringConfigs.items():
                string_configs[k] = v

        intConfigs = server_cfg.intConfigs if server_cfg.intConfigs is not None else {}
        if client_cfg.intConfigs is not None:
            for k,v in client_cfg.intConfigs.items():
                intConfigs[k] = v

        floatConfigs = server_cfg.floatConfigs if server_cfg.floatConfigs is not None else {}
        if client_cfg.floatConfigs is not None:
            for k,v in client_cfg.floatConfigs.items():
                floatConfigs[k] = v

        bytesConfigs = server_cfg.bytesConfigs if server_cfg.bytesConfigs is not None else {}
        if client_cfg.bytesConfigs is not None:
            for k,v in client_cfg.bytesConfigs.items():
                bytesConfigs[k] = v

        booleanConfigs = server_cfg.booleanConfigs if server_cfg.booleanConfigs is not None else {}
        if client_cfg.booleanConfigs is not None:
            for k,v in client_cfg.booleanConfigs.items():
                booleanConfigs[k] = v

        return common_pb2.IKVStoreConfig(stringConfigs=string_configs, intConfigs=intConfigs,\
                floatConfigs=floatConfigs, bytesConfigs=bytesConfigs, booleanConfigs=booleanConfigs)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63ommon.proto\x12\nikvschemas\"\x9b\x01\n\x11IKVDocumentOnWire\x12=\n\x08\x64ocument\x18\x01 \x03(\x0b\x32+.ikvschemas.IKVDocumentOnWire.DocumentEntry\x1aG\n\rDocumentEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.ikvschemas.FieldValue:\x02\x38\x01\"E\n\nFieldValue\x12(\n\tfieldType\x18\x01 \x01(\x0e\x32\x15.ikvschemas.FieldType\x12\r\n\x05value\x18\x02 \x01(\x0c\"\xf0\x04\n\x0eIKVStoreConfig\x12\x44\n\rstringConfigs\x18\x01 \x03(\x0b\x32-.ikvschemas.IKVStoreConfig.StringConfigsEntry\x12>\n\nintConfigs\x18\x02 \x03(\x0b\x32*.ikvschemas.IKVStoreConfig.IntConfigsEntry\x12\x42\n\x0c\x66loatConfigs\x18\x03 \x03(\x0b\x32,.ikvschemas.IKVStoreConfig.FloatConfigsEntry\x12\x42\n\x0c\x62ytesConfigs\x18\x04 \x03(\x0b\x32,.ikvschemas.IKVStoreConfig.BytesConfigsEntry\x12\x46\n\x0e\x62ooleanConfigs\x18\x05 \x03(\x0b\x32..ikvschemas.IKVStoreConfig.BooleanConfigsEntry\x1a\x34\n\x12StringConfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x31\n\x0fIntConfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\x1a\x33\n\x11\x46loatConfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a\x33\n\x11\x42ytesConfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x1a\x35\n\x13\x42ooleanConfigsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01*l\n\tFieldType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT64\x10\x02\x12\x0b\n\x07\x46LOAT32\x10\x03\x12\x0b\n\x07\x46LOAT64\x10\x04\x12\n\n\x06STRING\x10\x05\x12\t\n\x05\x42YTES\x10\x06\x12\x0b\n\x07\x42OOLEAN\x10\x07\x42 \n\x14\x63om.inlineio.schemasZ\x08schemas/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\024com.inlineio.schemasZ\010schemas/'
  _globals['_IKVDOCUMENTONWIRE_DOCUMENTENTRY']._options = None
  _globals['_IKVDOCUMENTONWIRE_DOCUMENTENTRY']._serialized_options = b'8\001'
  _globals['_IKVSTORECONFIG_STRINGCONFIGSENTRY']._options = None
  _globals['_IKVSTORECONFIG_STRINGCONFIGSENTRY']._serialized_options = b'8\001'
  _globals['_IKVSTORECONFIG_INTCONFIGSENTRY']._options = None
  _globals['_IKVSTORECONFIG_INTCONFIGSENTRY']._serialized_options = b'8\001'
  _globals['_IKVSTORECONFIG_FLOATCONFIGSENTRY']._options = None
  _globals['_IKVSTORECONFIG_FLOATCONFIGSENTRY']._serialized_options = b'8\001'
  _globals['_IKVSTORECONFIG_BYTESCONFIGSENTRY']._options = None
  _globals['_IKVSTORECONFIG_BYTESCONFIGSENTRY']._serialized_options = b'8\001'
  _globals['_IKVSTORECONFIG_BOOLEANCONFIGSENTRY']._options = None
  _globals['_IKVSTORECONFIG_BOOLEANCONFIGSENTRY']._serialized_options = b'8\001'
  _globals['_FIELDTYPE']._serialized_start=884
  _globals['_FIELDTYPE']._serialized_end=992
  _globals['_IKVDOCUMENTONWIRE']._serialized_start=29
  _globals['_IKVDOCUMENTONWIRE']._serialized_end=184
  _globals['_IKVDOCUMENTONWIRE_DOCUMENTENTRY']._serialized_start=113
  _globals['_IKVDOCUMENTONWIRE_DOCUMENTENTRY']._serialized_end=184
  _globals['_FIELDVALUE']._serialized_start=186
  _globals['_FIELDVALUE']._serialized_end=255
  _globals['_IKVSTORECONFIG']._serialized_start=258
  _globals['_IKVSTORECONFIG']._serialized_end=882
  _globals['_IKVSTORECONFIG_STRINGCONFIGSENTRY']._serialized_start=618
  _globals['_IKVSTORECONFIG_STRINGCONFIGSENTRY']._serialized_end=670
  _globals['_IKVSTORECONFIG_INTCONFIGSENTRY']._serialized_start=672
  _globals['_IKVSTORECONFIG_INTCONFIGSENTRY']._serialized_end=721
  _globals['_IKVSTORECONFIG_FLOATCONFIGSENTRY']._serialized_start=723
  _globals['_IKVSTORECONFIG_FLOATCONFIGSENTRY']._serialized_end=774
  _globals['_IKVSTORECONFIG_BYTESCONFIGSENTRY']._serialized_start=776
  _globals['_IKVSTORECONFIG_BYTESCONFIGSENTRY']._serialized_end=827
  _globals['_IKVSTORECONFIG_BOOLEANCONFIGSENTRY']._serialized_start=829
  _globals['_IKVSTORECONFIG_BOOLEANCONFIGSENTRY']._serialized_end=882
# @@protoc_insertion_point(module_scope)



from schemas.common_pb2 import IKVStoreConfig
import utils as ikvutils

class ClientOptions:
    """
    Reader/Writer client configuration options.
    Reference: https://docs.inlined.io/clients/python-guide#configuration
    See ClientOptionsBuilder to instantiate.
    """

    _VALID_LOG_LEVELS = {"error", "warn", "info", "debug", "trace"}

    def __init__(self, ikv_config: IKVStoreConfig):
        """ DO NOT use directly. See ClientOptionsBuilder to instantiate."""
        self._ikv_config = ikv_config
    
    def get_ikv_config(self) -> IKVStoreConfig:
        return self._ikv_config

class ClientOptionsBuilder:
    """
    Builder class for ClientOptions.
    Reference: https://docs.inlined.io/clients/python-guide#configuration
    """
    def __init__(self):
        ikv_config = IKVStoreConfig(stringConfigs={}, intConfigs={}, floatConfigs={}, bytesConfigs={}, booleanConfigs={})

        # default logging options
        ikv_config.stringConfigs["rust_client_log_level"] = "info"
        ikv_config.booleanConfigs["rust_client_log_to_console"] = True

        # TODO: remove explicit parititon
        ikv_config.intConfigs["partition"] = 0

        self.ikv_config: IKVStoreConfig = ikv_config
    
    def with_mount_directory(self, dir: str) -> 'ClientOptionsBuilder':
        self.ikv_config.stringConfigs["mount_directory"] = ikvutils.is_valid_str_or_raise(dir)
        return self

    def with_store_name(self, name: str) -> 'ClientOptionsBuilder':
        self.ikv_config.stringConfigs["store_name"] = ikvutils.is_valid_str_or_raise(name)
        return self

    def with_account_id(self, id: str) -> 'ClientOptionsBuilder':
        self.ikv_config.stringConfigs["account_id"] = ikvutils.is_valid_str_or_raise(id)
        return self

    def with_account_passkey(self, passkey: str) -> 'ClientOptionsBuilder':
        self.ikv_config.stringConfigs["account_passkey"] = ikvutils.is_valid_str_or_raise(passkey)
        return self

    def with_console_logging(self, level: str) -> 'ClientOptionsBuilder':
        level = ikvutils.is_valid_str_or_raise(level).lower()
        if not level in ClientOptions._VALID_LOG_LEVELS:
            raise ValueError("not a valid log level, valid levels: " + ", ".join(ClientOptions._VALID_LOG_LEVELS))

        self.ikv_config.stringConfigs["rust_client_log_level"] = level
        self.ikv_config.booleanConfigs["rust_client_log_to_console"] = True
        return self

    def with_file_logging(self, level: str, filepath: str) -> 'ClientOptionsBuilder':
        level = ikvutils.is_valid_str_or_raise(level).lower()
        if not level in ClientOptions._VALID_LOG_LEVELS:
            raise ValueError("not a valid log level, valid levels: " + ", ".join(ClientOptions._VALID_LOG_LEVELS))
        
        self.ikv_config.stringConfigs["rust_client_log_level"] = level
        self.ikv_config.booleanConfigs["rust_client_log_to_console"] = False
        self.ikv_config.stringConfigs["rust_client_log_file"] = ikvutils.is_valid_str_or_raise(filepath)
        return self

    def build(self) -> ClientOptions:
        return ClientOptions(self.ikv_config)


from typing import Optional
from ikvpy.utils import is_valid_str_or_raise


class EnvReader():
    def __init__(self) -> None:
        self.env = read_key_value_pairs(".env")
    
    def get(self, key: str) -> Optional[str]:
        return is_valid_str_or_raise(self.env[key])


def read_key_value_pairs(file_path):
    result = {}
    with open(file_path, 'r') as file:
        for line in file:
            line = line.strip()
            if '=' in line:
                key, value = line.split('=', 1)
                result[key.strip()] = value.strip()
    return result




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: services.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eservices.proto\x12\nikvschemas\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0c\x63ommon.proto\"!\n\x11HelloWorldRequest\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\"\"\n\x12HelloWorldResponse\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\"#\n\x13HealthCheckResponse\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\"\xc8\x01\n\x18UpsertFieldValuesRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x08\x64ocument\x18\x03 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"\xce\x01\n\x1d\x42\x61tchUpsertFieldValuesRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\tdocuments\x18\x03 \x03(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"\xdd\x01\n\x17\x44\x65leteFieldValueRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\ndocumentId\x18\x03 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\x12\x12\n\nfieldNames\x18\x04 \x03(\t\"\xe4\x01\n\x1d\x42\x61tchDeleteFieldValuesRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0b\x64ocumentIds\x18\x03 \x03(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\x12\x12\n\nfieldNames\x18\x04 \x03(\t\"\xc7\x01\n\x15\x44\x65leteDocumentRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\ndocumentId\x18\x03 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"\xce\x01\n\x1b\x42\x61tchDeleteDocumentsRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0b\x64ocumentIds\x18\x03 \x03(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"\xd4\x01\n\x11\x44ropFieldsRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x66ield_names\x18\x03 \x03(\t\x12\x1b\n\x13\x66ield_name_prefixes\x18\x04 \x03(\t\x12\x10\n\x08\x64rop_all\x18\x05 \x01(\x08\"Q\n\x0b\x46ieldSchema\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12(\n\tfieldType\x18\x03 \x01(\x0e\x32\x15.ikvschemas.FieldType\"i\n\x19GetUserStoreConfigRequest\x12L\n\x1buserStoreContextInitializer\x18\x01 \x01(\x0b\x32\'.ikvschemas.UserStoreContextInitializer\"N\n\x1aGetUserStoreConfigResponse\x12\x30\n\x0cglobalConfig\x18\x01 \x01(\x0b\x32\x1a.ikvschemas.IKVStoreConfig\"\x08\n\x06Status\"e\n\x1bUserStoreContextInitializer\x12\x33\n\x0b\x63redentials\x18\x01 \x01(\x0b\x32\x1e.ikvschemas.AccountCredentials\x12\x11\n\tstoreName\x18\x02 \x01(\t\"?\n\x12\x41\x63\x63ountCredentials\x12\x11\n\taccountId\x18\x01 \x01(\t\x12\x16\n\x0e\x61\x63\x63ountPasskey\x18\x02 \x01(\t2\x88\x06\n\x14InlineKVWriteService\x12M\n\nhelloWorld\x12\x1d.ikvschemas.HelloWorldRequest\x1a\x1e.ikvschemas.HelloWorldResponse\"\x00\x12O\n\x11upsertFieldValues\x12$.ikvschemas.UpsertFieldValuesRequest\x1a\x12.ikvschemas.Status\"\x00\x12Y\n\x16\x62\x61tchUpsertFieldValues\x12).ikvschemas.BatchUpsertFieldValuesRequest\x1a\x12.ikvschemas.Status\"\x00\x12N\n\x11\x64\x65leteFieldValues\x12#.ikvschemas.DeleteFieldValueRequest\x1a\x12.ikvschemas.Status\"\x00\x12Y\n\x16\x62\x61tchDeleteFieldValues\x12).ikvschemas.BatchDeleteFieldValuesRequest\x1a\x12.ikvschemas.Status\"\x00\x12I\n\x0e\x64\x65leteDocument\x12!.ikvschemas.DeleteDocumentRequest\x1a\x12.ikvschemas.Status\"\x00\x12U\n\x14\x62\x61tchDeleteDocuments\x12\'.ikvschemas.BatchDeleteDocumentsRequest\x1a\x12.ikvschemas.Status\"\x00\x12\x41\n\ndropFields\x12\x1d.ikvschemas.DropFieldsRequest\x1a\x12.ikvschemas.Status\"\x00\x12\x65\n\x12getUserStoreConfig\x12%.ikvschemas.GetUserStoreConfigRequest\x1a&.ikvschemas.GetUserStoreConfigResponse\"\x00\x32X\n\x0c\x41\x64minService\x12H\n\x0bhealthCheck\x12\x16.google.protobuf.Empty\x1a\x1f.ikvschemas.HealthCheckResponse\"\x00\x42 \n\x14\x63om.inlineio.schemasZ\x08schemas/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'services_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\024com.inlineio.schemasZ\010schemas/'
  _globals['_HELLOWORLDREQUEST']._serialized_start=106
  _globals['_HELLOWORLDREQUEST']._serialized_end=139
  _globals['_HELLOWORLDRESPONSE']._serialized_start=141
  _globals['_HELLOWORLDRESPONSE']._serialized_end=175
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=177
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=212
  _globals['_UPSERTFIELDVALUESREQUEST']._serialized_start=215
  _globals['_UPSERTFIELDVALUESREQUEST']._serialized_end=415
  _globals['_BATCHUPSERTFIELDVALUESREQUEST']._serialized_start=418
  _globals['_BATCHUPSERTFIELDVALUESREQUEST']._serialized_end=624
  _globals['_DELETEFIELDVALUEREQUEST']._serialized_start=627
  _globals['_DELETEFIELDVALUEREQUEST']._serialized_end=848
  _globals['_BATCHDELETEFIELDVALUESREQUEST']._serialized_start=851
  _globals['_BATCHDELETEFIELDVALUESREQUEST']._serialized_end=1079
  _globals['_DELETEDOCUMENTREQUEST']._serialized_start=1082
  _globals['_DELETEDOCUMENTREQUEST']._serialized_end=1281
  _globals['_BATCHDELETEDOCUMENTSREQUEST']._serialized_start=1284
  _globals['_BATCHDELETEDOCUMENTSREQUEST']._serialized_end=1490
  _globals['_DROPFIELDSREQUEST']._serialized_start=1493
  _globals['_DROPFIELDSREQUEST']._serialized_end=1705
  _globals['_FIELDSCHEMA']._serialized_start=1707
  _globals['_FIELDSCHEMA']._serialized_end=1788
  _globals['_GETUSERSTORECONFIGREQUEST']._serialized_start=1790
  _globals['_GETUSERSTORECONFIGREQUEST']._serialized_end=1895
  _globals['_GETUSERSTORECONFIGRESPONSE']._serialized_start=1897
  _globals['_GETUSERSTORECONFIGRESPONSE']._serialized_end=1975
  _globals['_STATUS']._serialized_start=1977
  _globals['_STATUS']._serialized_end=1985
  _globals['_USERSTORECONTEXTINITIALIZER']._serialized_start=1987
  _globals['_USERSTORECONTEXTINITIALIZER']._serialized_end=2088
  _globals['_ACCOUNTCREDENTIALS']._serialized_start=2090
  _globals['_ACCOUNTCREDENTIALS']._serialized_end=2153
  _globals['_INLINEKVWRITESERVICE']._serialized_start=2156
  _globals['_INLINEKVWRITESERVICE']._serialized_end=2932
  _globals['_ADMINSERVICE']._serialized_start=2934
  _globals['_ADMINSERVICE']._serialized_end=3022
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class EventHeader(_message.Message):
    __slots__ = ("sourceTimestamp",)
    SOURCETIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    sourceTimestamp: _timestamp_pb2.Timestamp
    def __init__(self, sourceTimestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class IKVDataEvent(_message.Message):
    __slots__ = ("eventHeader", "upsertDocumentFieldsEvent", "deleteDocumentFieldsEvent", "deleteDocumentEvent", "dropFieldEvent")
    EVENTHEADER_FIELD_NUMBER: _ClassVar[int]
    UPSERTDOCUMENTFIELDSEVENT_FIELD_NUMBER: _ClassVar[int]
    DELETEDOCUMENTFIELDSEVENT_FIELD_NUMBER: _ClassVar[int]
    DELETEDOCUMENTEVENT_FIELD_NUMBER: _ClassVar[int]
    DROPFIELDEVENT_FIELD_NUMBER: _ClassVar[int]
    eventHeader: EventHeader
    upsertDocumentFieldsEvent: UpsertDocumentFieldsEvent
    deleteDocumentFieldsEvent: DeleteDocumentFieldsEvent
    deleteDocumentEvent: DeleteDocumentEvent
    dropFieldEvent: DropFieldEvent
    def __init__(self, eventHeader: _Optional[_Union[EventHeader, _Mapping]] = ..., upsertDocumentFieldsEvent: _Optional[_Union[UpsertDocumentFieldsEvent, _Mapping]] = ..., deleteDocumentFieldsEvent: _Optional[_Union[DeleteDocumentFieldsEvent, _Mapping]] = ..., deleteDocumentEvent: _Optional[_Union[DeleteDocumentEvent, _Mapping]] = ..., dropFieldEvent: _Optional[_Union[DropFieldEvent, _Mapping]] = ...) -> None: ...

class UpsertDocumentFieldsEvent(_message.Message):
    __slots__ = ("document",)
    DOCUMENT_FIELD_NUMBER: _ClassVar[int]
    document: _common_pb2.IKVDocumentOnWire
    def __init__(self, document: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ...) -> None: ...

class DeleteDocumentFieldsEvent(_message.Message):
    __slots__ = ("documentId", "fieldsToDelete")
    DOCUMENTID_FIELD_NUMBER: _ClassVar[int]
    FIELDSTODELETE_FIELD_NUMBER: _ClassVar[int]
    documentId: _common_pb2.IKVDocumentOnWire
    fieldsToDelete: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, documentId: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ..., fieldsToDelete: _Optional[_Iterable[str]] = ...) -> None: ...

class DeleteDocumentEvent(_message.Message):
    __slots__ = ("documentId",)
    DOCUMENTID_FIELD_NUMBER: _ClassVar[int]
    documentId: _common_pb2.IKVDocumentOnWire
    def __init__(self, documentId: _Optional[_Union[_common_pb2.IKVDocumentOnWire, _Mapping]] = ...) -> None: ...

class DropFieldEvent(_message.Message):
    __slots__ = ("field_names", "field_name_prefixes", "drop_all")
    FIELD_NAMES_FIELD_NUMBER: _ClassVar[int]
    FIELD_NAME_PREFIXES_FIELD_NUMBER: _ClassVar[int]
    DROP_ALL_FIELD_NUMBER: _ClassVar[int]
    field_names: _containers.RepeatedScalarFieldContainer[str]
    field_name_prefixes: _containers.RepeatedScalarFieldContainer[str]
    drop_all: bool
    def __init__(self, field_names: _Optional[_Iterable[str]] = ..., field_name_prefixes: _Optional[_Iterable[str]] = ..., drop_all: bool = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: streaming.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fstreaming.proto\x12\nikvschemas\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x0c\x63ommon.proto\"[\n\x0b\x45ventHeader\x12\x38\n\x0fsourceTimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x42\x12\n\x10_sourceTimestamp\"\xd3\x02\n\x0cIKVDataEvent\x12,\n\x0b\x65ventHeader\x18\x01 \x01(\x0b\x32\x17.ikvschemas.EventHeader\x12J\n\x19upsertDocumentFieldsEvent\x18\x02 \x01(\x0b\x32%.ikvschemas.UpsertDocumentFieldsEventH\x00\x12J\n\x19\x64\x65leteDocumentFieldsEvent\x18\x03 \x01(\x0b\x32%.ikvschemas.DeleteDocumentFieldsEventH\x00\x12>\n\x13\x64\x65leteDocumentEvent\x18\x04 \x01(\x0b\x32\x1f.ikvschemas.DeleteDocumentEventH\x00\x12\x34\n\x0e\x64ropFieldEvent\x18\x05 \x01(\x0b\x32\x1a.ikvschemas.DropFieldEventH\x00\x42\x07\n\x05\x65vent\"L\n\x19UpsertDocumentFieldsEvent\x12/\n\x08\x64ocument\x18\x01 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"f\n\x19\x44\x65leteDocumentFieldsEvent\x12\x31\n\ndocumentId\x18\x01 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\x12\x16\n\x0e\x66ieldsToDelete\x18\x02 \x03(\t\"H\n\x13\x44\x65leteDocumentEvent\x12\x31\n\ndocumentId\x18\x01 \x01(\x0b\x32\x1d.ikvschemas.IKVDocumentOnWire\"T\n\x0e\x44ropFieldEvent\x12\x13\n\x0b\x66ield_names\x18\x01 \x03(\t\x12\x1b\n\x13\x66ield_name_prefixes\x18\x02 \x03(\t\x12\x10\n\x08\x64rop_all\x18\x03 \x01(\x08\x42 \n\x14\x63om.inlineio.schemasZ\x08schemas/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'streaming_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\024com.inlineio.schemasZ\010schemas/'
  _globals['_EVENTHEADER']._serialized_start=78
  _globals['_EVENTHEADER']._serialized_end=169
  _globals['_IKVDATAEVENT']._serialized_start=172
  _globals['_IKVDATAEVENT']._serialized_end=511
  _globals['_UPSERTDOCUMENTFIELDSEVENT']._serialized_start=513
  _globals['_UPSERTDOCUMENTFIELDSEVENT']._serialized_end=589
  _globals['_DELETEDOCUMENTFIELDSEVENT']._serialized_start=591
  _globals['_DELETEDOCUMENTFIELDSEVENT']._serialized_end=693
  _globals['_DELETEDOCUMENTEVENT']._serialized_start=695
  _globals['_DELETEDOCUMENTEVENT']._serialized_end=767
  _globals['_DROPFIELDEVENT']._serialized_start=769
  _globals['_DROPFIELDEVENT']._serialized_end=853
# @@protoc_insertion_point(module_scope)



import unittest
from src.ikvpy.native_reader import NativeReader

class TestNativeReader(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.native_reader = \
            NativeReader("/Users/pushkar/projects/ikv-store/ikv/target/release/libikv.dylib")

    def test_health_check(self):
        self.assertTrue(self.native_reader.health_check("foo"))

# run: $> python3 tests/test_native_reader.py
#      $> [ikv-python-client/tests]$ python3 -m unittest
if __name__ == '__main__':
    unittest.main()


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import services_pb2 as services__pb2


class InlineKVWriteServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.helloWorld = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/helloWorld',
                request_serializer=services__pb2.HelloWorldRequest.SerializeToString,
                response_deserializer=services__pb2.HelloWorldResponse.FromString,
                )
        self.upsertFieldValues = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/upsertFieldValues',
                request_serializer=services__pb2.UpsertFieldValuesRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.batchUpsertFieldValues = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/batchUpsertFieldValues',
                request_serializer=services__pb2.BatchUpsertFieldValuesRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.deleteFieldValues = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/deleteFieldValues',
                request_serializer=services__pb2.DeleteFieldValueRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.batchDeleteFieldValues = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/batchDeleteFieldValues',
                request_serializer=services__pb2.BatchDeleteFieldValuesRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.deleteDocument = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/deleteDocument',
                request_serializer=services__pb2.DeleteDocumentRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.batchDeleteDocuments = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/batchDeleteDocuments',
                request_serializer=services__pb2.BatchDeleteDocumentsRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.dropFields = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/dropFields',
                request_serializer=services__pb2.DropFieldsRequest.SerializeToString,
                response_deserializer=services__pb2.Status.FromString,
                )
        self.getUserStoreConfig = channel.unary_unary(
                '/ikvschemas.InlineKVWriteService/getUserStoreConfig',
                request_serializer=services__pb2.GetUserStoreConfigRequest.SerializeToString,
                response_deserializer=services__pb2.GetUserStoreConfigResponse.FromString,
                )


class InlineKVWriteServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def helloWorld(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def upsertFieldValues(self, request, context):
        """Write methods
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def batchUpsertFieldValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteFieldValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def batchDeleteFieldValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteDocument(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def batchDeleteDocuments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def dropFields(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getUserStoreConfig(self, request, context):
        """Gateway-specified configuration
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_InlineKVWriteServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'helloWorld': grpc.unary_unary_rpc_method_handler(
                    servicer.helloWorld,
                    request_deserializer=services__pb2.HelloWorldRequest.FromString,
                    response_serializer=services__pb2.HelloWorldResponse.SerializeToString,
            ),
            'upsertFieldValues': grpc.unary_unary_rpc_method_handler(
                    servicer.upsertFieldValues,
                    request_deserializer=services__pb2.UpsertFieldValuesRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'batchUpsertFieldValues': grpc.unary_unary_rpc_method_handler(
                    servicer.batchUpsertFieldValues,
                    request_deserializer=services__pb2.BatchUpsertFieldValuesRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'deleteFieldValues': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteFieldValues,
                    request_deserializer=services__pb2.DeleteFieldValueRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'batchDeleteFieldValues': grpc.unary_unary_rpc_method_handler(
                    servicer.batchDeleteFieldValues,
                    request_deserializer=services__pb2.BatchDeleteFieldValuesRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'deleteDocument': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteDocument,
                    request_deserializer=services__pb2.DeleteDocumentRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'batchDeleteDocuments': grpc.unary_unary_rpc_method_handler(
                    servicer.batchDeleteDocuments,
                    request_deserializer=services__pb2.BatchDeleteDocumentsRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'dropFields': grpc.unary_unary_rpc_method_handler(
                    servicer.dropFields,
                    request_deserializer=services__pb2.DropFieldsRequest.FromString,
                    response_serializer=services__pb2.Status.SerializeToString,
            ),
            'getUserStoreConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.getUserStoreConfig,
                    request_deserializer=services__pb2.GetUserStoreConfigRequest.FromString,
                    response_serializer=services__pb2.GetUserStoreConfigResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ikvschemas.InlineKVWriteService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class InlineKVWriteService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def helloWorld(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/helloWorld',
            services__pb2.HelloWorldRequest.SerializeToString,
            services__pb2.HelloWorldResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def upsertFieldValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/upsertFieldValues',
            services__pb2.UpsertFieldValuesRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def batchUpsertFieldValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/batchUpsertFieldValues',
            services__pb2.BatchUpsertFieldValuesRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteFieldValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/deleteFieldValues',
            services__pb2.DeleteFieldValueRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def batchDeleteFieldValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/batchDeleteFieldValues',
            services__pb2.BatchDeleteFieldValuesRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteDocument(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/deleteDocument',
            services__pb2.DeleteDocumentRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def batchDeleteDocuments(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/batchDeleteDocuments',
            services__pb2.BatchDeleteDocumentsRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def dropFields(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/dropFields',
            services__pb2.DropFieldsRequest.SerializeToString,
            services__pb2.Status.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getUserStoreConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.InlineKVWriteService/getUserStoreConfig',
            services__pb2.GetUserStoreConfigRequest.SerializeToString,
            services__pb2.GetUserStoreConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class AdminServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.healthCheck = channel.unary_unary(
                '/ikvschemas.AdminService/healthCheck',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=services__pb2.HealthCheckResponse.FromString,
                )


class AdminServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def healthCheck(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdminServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'healthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.healthCheck,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=services__pb2.HealthCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ikvschemas.AdminService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AdminService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def healthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ikvschemas.AdminService/healthCheck',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            services__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import time
import unittest
from src.ikvpy.clientoptions import ClientOptions, ClientOptionsBuilder
import src.ikvpy.client as ikv_client
import src.ikvpy.document as ikv_document
from src.ikvpy.factory import create_new_reader, create_new_writer
from env_var import EnvReader

class TestInteg(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # setup client options
        env = EnvReader()
        client_options: ClientOptions = ClientOptionsBuilder() \
            .with_account_id(env.get("ACCOUNT_ID"))\
            .with_account_passkey(env.get("ACCOUNT_PASSKEY"))\
            .with_store_name(env.get("STORE_NAME"))\
            .with_mount_directory(env.get("MOUNT_DIR"))\
            .build()
        self.reader: ikv_client.IKVReader = create_new_reader(client_options)
        self.writer: ikv_client.IKVWriter = create_new_writer(client_options)
        self.reader.startup()
        self.writer.startup()

    @classmethod
    def tearDownClass(self):
        self.reader.shutdown()
        self.writer.shutdown()

    def test_single_doc(self):
        # upsert {"userid": "0", "firstname": "Alice"}
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "0")\
            .put_string_field("firstname", "Alice").build()
        self.writer.upsert_fields(document)
        time.sleep(5)

        # read valid fields
        self.assertEqual("0", self.reader.get_string_value("0", "userid"))
        self.assertEqual("Alice", self.reader.get_string_value("0", "firstname"))

        # read valid fields for invalid document
        self.assertIsNone(self.reader.get_string_value("1000", "userid"))

        # read invalid fields for valid document
        self.assertIsNone(self.reader.get_string_value("0", "foo_field"))
        
    def test_multiget(self):
        # empty check
        empty_list = list(self.reader.multiget_string_values(field_names=["firstname"]))
        self.assertEqual(len(empty_list), 0)

        # upsert {"userid": "8", "firstname": "Alice"}
        # upsert {"userid": "9", "firstname": "Bob", "city": "NYC"}
        # upsert {"userid": "10", "firstname": "Allison", "city": "NYC"}
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "8")\
            .put_string_field("firstname", "Alice").build()
        self.writer.upsert_fields(document)
        
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "9")\
            .put_string_field("firstname", "Bob").put_string_field("city", "NYC").build()
        self.writer.upsert_fields(document)
        
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "10")\
            .put_string_field("firstname", "Allison").put_string_field("city", "NYC").build()
        self.writer.upsert_fields(document)

        time.sleep(5)

        # get for "10" and "9"
        multiget_iter = self.reader.multiget_string_values(bytes_primary_keys=["10".encode("utf-8")],
            str_primary_keys=["8", "9", "foo"], field_names=["firstname", "bar", "city"])
        drained_multiget_iter = list(multiget_iter)
        
        # ouput:
        # 10#firstname,None,10#city
        # 8#firstname,None,8#city
        # 9#firstname,None,9#city
        # None,None,None
        self.assertEqual(["Allison",None,"NYC","Alice",None,None,"Bob",None,"NYC",None,None,None],
            drained_multiget_iter)

    def test_upsert_delete_doc(self):
        # upsert {"userid": "1", "firstname": "Alice"}
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "1")\
            .put_string_field("firstname", "Alice").build()
        self.writer.upsert_fields(document)
       
        # delete {"userid": "1"}
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "1").build()
        self.writer.delete_document(document)

        time.sleep(5)

        # read assertions
        self.assertIsNone(self.reader.get_string_value("1", "userid"))
        self.assertIsNone(self.reader.get_string_value("1", "firstname"))

    def test_upsert_drop_fields(self):
        # upsert {"userid": "5", "firstname": "Alice"}
        # upsert {"userid": "6", "firstname": "Bob", "city": "NYC"}
        # upsert {"userid": "7", "firstname": "Allison", "city": "NYC"}
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "5")\
            .put_string_field("firstname", "Alice").build()
        self.writer.upsert_fields(document)
        
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "6")\
            .put_string_field("firstname", "Bob").put_string_field("city", "NYC").build()
        self.writer.upsert_fields(document)
        
        document = ikv_document.IKVDocumentBuilder().put_string_field("userid", "7")\
            .put_string_field("firstname", "Allison").put_string_field("city", "NYC").build()
        self.writer.upsert_fields(document)

        # drop column for "firstname" and "city"
        self.writer.drop_fields_by_name(["firstname", "foo"])
        self.writer.drop_fields_by_name_prefix(["ci", "ciii"])

        time.sleep(5)

        # read assertions
        self.assertEqual("5", self.reader.get_string_value("5", "userid"))
        self.assertEqual("6", self.reader.get_string_value("6", "userid"))
        self.assertEqual("7", self.reader.get_string_value("7", "userid"))
        
        self.assertIsNone(self.reader.get_string_value("5", "firstname"))
        self.assertIsNone(self.reader.get_string_value("6", "firstname"))
        self.assertIsNone(self.reader.get_string_value("7", "firstname"))
        
        self.assertIsNone(self.reader.get_string_value("5", "city"))
        self.assertIsNone(self.reader.get_string_value("6", "city"))
        self.assertIsNone(self.reader.get_string_value("7", "city"))

# run: $> python3 tests/test_integ.py
#      $> [ikv-python-client/tests]$ python3 -m unittest
if __name__ == '__main__':
    unittest.main()


from clientoptions import ClientOptions
from reader import IKVReaderImpl
from writer import IKVWriterImpl
from client import IKVReader, IKVWriter

"""
Factory for creating concrete IKVReader and IKVWriter instances.
"""

def create_new_reader(client_options: ClientOptions) -> IKVReader:
    """
    Create new reader client by supplying configuration.
    args:
        client_options: reader configuration, see clientoptions.py to instantiate
    """
    return IKVReaderImpl(client_options)

def create_new_writer(client_options: ClientOptions) -> IKVWriter:
    """
    Create new writer client by supplying configuration.
    args:
        client_options: writer configuration, see clientoptions.py to instantiate
    """
    return IKVWriterImpl(client_options)